\ grace.fth is the actions file for the Grace parser generator
\ It works with gracegen.fth that generates a parser in Forth source code

\ Copyright (C) Gerry Jackson 2008

\ This software is free; you can redistribute it and/or modify it in
\ any way provided you acknowledge the original source and copyright

\ This program is distributed in the hope that it will be useful,
\ but WITHOUT ANY WARRANTY; without even the implied warranty of
\ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

\ -----------------------------------------------------------------------------

.( Loading grace.fth ... ) cr

\ ---[ Optional screen display ]------------------------------------------------
\ Depends on user settable flag verbose

: ?cr    ( -- ) verbose if cr then ;
: ?space ( -- ) verbose if space then ;
: ?type  ( caddr u -- ) verbose if type else 2drop then ;

\ ---[ Interface to the scanner ]-----------------------------------------------

BaseDefaultData #states CheckNextData maxCheck init-scanner value bnf-scanner

: getNextToken ( -- caddr u tok )
   begin
      bnf-scanner next-token dup white_space =
   while
      ?space drop 2drop
   repeat
;

\ ---[ BNF source file handling  ]----------------------------------------------

0 value bnf-source

: open-bnf-source
   r/o open-file throw
   file-source dup set-source
   to bnf-source
;

: close-bnf-source  ( -- )
   bnf-scanner close-scanner
   source-id close-file throw
   set-ANS-source
   bnf-source close-source 0 to bnf-source
;

\ ---[ Shared variables ]-------------------------------------------------------

variable sym
2variable symname
32 chars constant ident-size   \ Maximum identifier length
create ident-buf ident-size allot align
variable ident-len

\ ---[ Interface to the BNF parser ]--------------------------------------------

\ Values to be loaded by the parser

0 value first-set
0 value bytes/set
0 value bits/cell
0 value parser-name

tokenval @ 1- max-member  \ The set size

\ Test if sym has its bit number set in the bit vector
\ This version for sets <= 1 cell in size

: testsym? ( set-index -- f )
   bytes/set * first-set + @     ( -- vec )
   1 sym @ lshift and
;

: test-token  ( n -- f )  sym @ = ;

: reportError  ( -- )
   source                           ( -- caddr u )
   cr over swap type cr             ( -- caddr )
   symname 2@ drop                  ( -- caddr caddr2 )
   swap - spaces ." ^ "
   -1 abort" syntax error"
;

: >ident-string  ( caddr u -- )
   ident-buf swap ident-size min
   dup ident-len ! move
;

: ?nextsym  ( f -- )
   0= if reportError close-bnf-source then
   begin
      getNextToken dup 0<     ( -- caddr u tok f )
   while
      drop 2drop refill 0=    ( -- f2 )
   until
      source eof_token
   then
   dup ident = over terminal_string = or
   if >r 2dup >ident-string r> then
   sym ! 2dup ?type symname 2!
;

\ ---[ Actions embedded in the grammar and helpers ]---------------------------

: ident-string ( -- caddr u )
   ident-buf ident-len @
;

: startGrammar ( -- )  ?cr ;

 
variable current-nt-def

\ Because an ident can be either a terminal or non-terminal startRule checks
\ that syntax-expr is a non-terminal

: startRule  ( syntax-expr -- )
   dup @ nt-Syntax-Expr <>       \ @ gets the class
   if cr .in ." Invalid name for rule" gray-error then
   current-nt-def ! ((
;

: endRule  ( -- syntax-expr )
   )) current-nt-def @ rule
   ?cr
;

: saveQuote  ( caddr -- )  [char] " swap c! ;

: make-quoted  ( caddr u -- )
   over saveQuote
   1- chars + saveQuote
;

: str>countedstr  ( caddr u -- ad )  \ Convert to a counted string at HERE
   here 2dup c!               ( -- caddr u ad )
   char+ swap move
   here
;

\ Wordlists for terminals and non-terminals

wordlist constant terminal-wordlist
wordlist constant nonterminal-wordlist

: getTerminal ( -- syntax-expr )
   ident-string 2dup make-quoted    \ Ensure in the form "name" before search
   terminal-wordlist search-wordlist
   0= abort" Terminal not declared"
   execute
;

\ Because identifiers can be either a terminal or non-terminal, find-nonterm
\ must search the terminal-wordlist first as all terminals will already have
\ been declared

: find-nonterm  ( -- 0 | xt f )
   ident-string terminal-wordlist search-wordlist ?dup 0=
   if ident-string nonterminal-wordlist search-wordlist then
;

create evalbuf 64 allot    \ Buffer to hold nonterminal <name> for evaluation

: string>buf  ( u2 caddr u -- u3 )  \ u2 & u3 are before/after evalbuf offsets
   rot 2dup + >r              ( -- caddr u u2 )
   evalbuf + swap move r>     ( -- u3 )
;

\ Evaluates "nonterminal <name>" to declare a non-terminal

: declare-nt  ( -- xt )
   get-current >r nonterminal-wordlist set-current
   0 s" nonterminal " string>buf
   ident-string string>buf evalbuf swap evaluate
   find-nonterm
   0= abort" Non-terminal declaration failed"
   r> set-current
;

2variable grammar-name  0 grammar-name !

\ The first non-terminal provides the grammar name

: ?save-grammar-name  ( -- )
   grammar-name @ 0=
   if
      ident-string here over     ( -- caddr u ad u )
      2dup grammar-name 2! allot ( -- caddr u ad )
      swap move                  ( -- )
   then
;

: getNonTerminal ( -- syntax-expr )
   ?save-grammar-name
   find-nonterm 0=
   if declare-nt then
   execute                    ( -- syntax-expr )
;

: skipInput  ( del -- )
   begin
      true ?nextsym sym @ over <>
   while
      sym @ eof_token =
      if
         close-bnf-source cr -1
         abort" Premature end of file - comment delimiter not found"
      then
   repeat
   drop symname 2@ drop source drop - >in !
;

: skipComment  ( -- )  sym @ "*/" skipInput sym ! ?cr ;

: endAction  ( -- )  ;  \ Nothing to do ??

: endGrammar  ( -- )
   grammar-name 2@ nonterminal-wordlist search-wordlist
   if
      execute                          ( -- syntax-expr )
   else
      -1 abort" Error - grammar name not found in non-terminals"
   then
;

1 tokenval !   \ Reset for the file to be compiled

\ ---[ To parse a BNF file ]----------------------------------------------------

\ grace runs the parser on the file given by (caddr u)

: grace ( caddr u -- syntax-expr )     \ (caddr u) is the .bnf source file
   ?cr open-bnf-source refill
   if
      true ?nextsym parser-name execute
   else
      cr ." Empty BNF file"
   then
   close-bnf-source
;

\ -----------------------------------------------------------------------------


.( grace.fth loaded. ) .s
