\ The concrete generator classes to generate the parser as Forth source code

\ Copyright (C) Gerry Jackson 2008

\ This software is free; you can redistribute it and/or modify it in
\ any way provided you acknowledge the original source and copyright

\ This program is distributed in the hope that it will be useful,
\ but WITHOUT ANY WARRANTY; without even the implied warranty of
\ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

\ -----------------------------------------------------------------------------

.( Loading gracegen.fth ... ) cr

\ ---[ Sections for non-terminal buffer and actions ]--------------------------

dataspace get-section value ds
: ForthDataSpace ds restore-section ;

dataspace 14000 section construct nonterm-buf  \ Increase for big non-terminals
dataspace 11000 section construct action-buf   \ Increase for large grammar

ForthDataSpace

\ Section primitives

: text>buf  ( caddr u -- )  here over allot swap move ;

: string>buf  ( caddr u  -- )  dup if dup c, text>buf else 2drop then ;

: char>buf  ( ch -- ) 1 c, c, ;

: newLine>buf  ( -- )  1 c, ( 13 c, ) 10 c, ;

255 constant ws-flag
128 constant max-target-llen     \ Maximum unformatted line length 

: write-buf  ( ad -- )
   begin
      dup c@
   while
      dup count dup ws-flag =       ( -- ad caddr u f ) 
      if
         drop c@ writeSpaces
         2 chars +                  ( -- ad' )
      else
         tuck writeText             ( -- ad u )
         1+ chars +                 ( -- ad' )
      then
   repeat
   drop
;

\ ---[ White space handling ]--------------------------------------------------
\ To make the generated code readable

variable tabs 0 tabs !
3 constant tabk
39 constant maxtab  \ Give up tabbing at maxtab to avoid overlong lines 

: +tab  tabk tabs +! ;
: -tab  tabs @ tabk - 0 max tabs ! ;    \ Ensure it never goes negative
: clear-tab  0 tabs ! ;
: get-tabs  tabs @ maxtab min ;

variable newline-flag true newline-flag !   \ Flag to indicate on a new line
variable llen  0 llen !    \ Cumulative length of the current line
80 constant max-llen

: ?llen  ( u -- f ) llen @ + max-llen > ;  \ Returns true if line too long

: clear-newline false newline-flag ! ;
: save-newline  newLine>buf true newline-flag ! 0 llen ! ;
: ?save-newline newline-flag @ 0= if save-newLine then ;

\ Save spaces in the buffer using run length encoding. Uses 2 bytes, the first
\ is ws-flag = 255 as a space indicator, the second the number of spaces.

: save-spaces  ( n -- )    \ 0 <= n < 256
   0 max ?dup
   if ws-flag c, c, then 
;

: save-tabs  get-tabs save-spaces ;

: save-tabs/space  ( -- )
   newline-flag @
   if
      save-tabs get-tabs
   else
      bl char>buf 1
   then
   llen +!
;

\ ---[ Save text in buffer ]---------------------------------------------------

: save-text  ( caddr u -- )
   dup llen +! string>buf clear-newline
;

: save-spaced-text  ( caddr u -- )  \ Saves 1 or #tab spaces before the text
   save-tabs/space save-text
;   

\ save-string is the word to call to save text in the buffer. This is done in
\ 3 stages:
\ - if not at the start of a new line it saves the string if there is
\   sufficient space in the rest of the line, else it saves a new line
\ - if at the start of a line the string length + tabs fits onto a line it
\   saves the tabs (spaces) and the string
\ - otherwise it saves the string without preceding spaces

: save-string  ( caddr u -- )
   newline-flag @ 0=
   if
      dup 1+ ?llen 0=
      if save-spaced-text exit then
      save-newline
   then
   dup get-tabs + ?llen
   if
      dup max-target-llen >= abort" Actions: line is too long"
      save-text
   else
      save-spaced-text
   then
;

: save-number  ( n -- )
   dup >r abs 0 <# #s r> sign #>
   save-string
;

: save-test  ( fs-index -- )
   save-number s" testsym?" save-string
;

: save-test-token  ( tok -- )     \ Test tok = symbol
   save-number s" test-token" save-string
; 

: save-check&next  ( -- )
   check&next-name dup 0=
   if .in -1 abort" check&next name has not been set" then
   save-string
;

: save-end-control  ( caddr u ) ?save-newline save-string save-newline ;
: save-control  ( caddr u ) save-end-control +tab ;
: save-defname  ( caddr u ) save-string save-newline +tab ;

\ ---[ Save Forth words in buffer ]--------------------------------------------

: save-if      ( -- )       s" if"     save-control ;
: save-else    ( -- )  -tab s" else"   save-control ;
: save-then    ( -- )  -tab s" then"   save-end-control ;
: save-begin   ( -- )       s" begin"  save-control ;
: save-while   ( -- )  -tab s" while"  save-control ;
: save-repeat  ( -- )  -tab s" repeat" save-end-control ;
: save-until   ( -- )  -tab s" until"  save-end-control ;
: save-:noname ( -- )  save-newline  clear-tab s" :noname" save-control ;
: save-:       ( -- )  save-newline  clear-tab s" :" save-text ;
: save-;       ( -- )  ?save-newline clear-tab s" ;" save-text ;
: save-0=      ( -- )  s" 0="      save-string ;
: save-and     ( -- )  s" and"     save-string ;
: save-is      ( -- )  s" is"      save-string ;
: save-recurse ( -- )  s" recurse" save-string ;

\ ---[ Preamble at start of output file ]--------------------------------------

: preamble
   s" \ Parser generated by Grace" writeLine writeNewLine
   empty get-set-size cells
   writeNumber s"  to bytes/set" writeLine
   bits/cell writeNumber s"  to bits/cell" writeLine
;

\ ---[ List of first sets ]----------------------------------------------------
\ A list of first-sets is built up so that they can be saved after all
\ code is generated. The order in the list will correspond to an index number
\ saved in each syntax-expr. Singleton sets will not be included as they
\ should have been filtered out earlier when "n test-token" will have been
\ saved to do an integer comparison as opposed to set membership (this is an
\ optimisation compared to the original Gray).

ListItem class
   cell var ^set
end-class SetListItem

:noname  ( set class -- obj )
   [ ListItem :: new ]
   tuck ^set !
; SetListItem defines new

List new constant first-set-list

: append-first-set  ( set -- )
   SetListItem new               ( -- sli )
   first-set-list list-append
;

\ compare-item is used by list-find to compare membership of the 2 sets

:noname  ( set item -- set f )   \ Returns false if equal
   ^set @ over equal? 0=
; SetListItem defines compare-item

\ get-fs-index returns the position of the supplied set in the first-set list
\ It searches the list, if not found the set is appended to the list. In both
\ cases its position in the list is returned

: get-fs-index  ( set -- pos )
   first-set-list list-find ?dup       ( -- set item item | set 0 )
   if
      first-set-list get-item-position ( -- set pos )
      nip
   else
      first-set-list list-length       ( -- set pos )
      swap append-first-set            ( -- pos )
   then
;

: save-vecs  ( vecn...vec1 n -- )
   hex 0
   ?do
      bits/cell 4 /
      writeNumberU.R s"  , " writeText
   loop
   decimal
;

: save-first-sets  ( -- )
   clear-tab writeNewLine
   s" here to first-set" writeLine
   s" base @ hex" writeLine
   first-set-list get-head
   begin
      ?dup
   while
      dup ^set @ get-bit-vector  ( -- set vecn...vec1 n )
      save-vecs writeNewLine     ( -- set )
      next-item @
   repeat
   s" base !" writeLine
;

: show-fsl     \ display first-set list
   cr first-set-list get-head
   begin
      ?dup
   while
      dup ^set @ . cr
      next-item @
   repeat
;

\ ---[ Common base class definitions ]-----------------------------------------

2variable current-nt 0 0 current-nt 2!   \ A global - be nice to eliminate this

\ begin-def and end-def are needed for all syntax-expr's as the user may 
\ have any of these on the stack when parser is executed

:noname  ( gen-class -- )
   get-section >r nonterm-buf
   drop 0 org
   save-: parse-name 2dup save-defname
   current-nt 2!
   r> restore-section
; dup GeneratorBase defines begin-def
      ActionGeneratorBase defines begin-def

:noname  ( gen-class -- )
   get-section >r nonterm-buf
   drop save-; save-newline 0 c,
   0 org here write-buf
   writeNewLine s" ' " writeText
   current-nt 2@ writeText
   s"  to parser-name" writeLine
   r> restore-section
; dup GeneratorBase defines end-def
  dup ActionGeneratorBase defines end-def
      Pass2GeneratorBase defines end-def

\ ---[ Concrete Generator classes ]--------------------------------------------

\ ---[ eps generator ]---------------------------------------------------------

GeneratorBase class
end-class Forth-Eps-Generator

' no-syntax-expr-op Forth-Eps-Generator defines generate()

Forth-Eps-Generator Eps-Syntax-Expr set-generator
 
\ ---[ Terminal generator ]----------------------------------------------------

GeneratorBase class
end-class Forth-Terminal-Generator

\ first-set-test generates code to test if a symbol is in the first-set. If
\ the set is a singleton it optimises it to test the token, otherwise
\ it tests the set.

: first-set-test  ( syntax-expr -- )
   get-first dup count-members 1 =
   if
      0 get-next-member             ( -- n )
      dup save-test-token
   else
      get-fs-index dup save-test    ( -- n1 )
   then
   set-index !                      \ May not need set-index field
;

:noname  ( gen-class -- )
   drop get-section >r nonterm-buf
   this first-set-test
   save-check&next
   r> restore-section
; Forth-Terminal-Generator defines generate()

Forth-Terminal-Generator Terminal-Syntax-Expr set-generator

\ ---[ Concatenation generator ]-----------------------------------------------

GeneratorBase class
end-class Forth-Concat-Generator

:noname  ( gen-class -- )
   drop
   operand1 generate
   operand2 generate
; Forth-Concat-Generator defines generate()

Forth-Concat-Generator Concat-Syntax-Expr set-generator

\ ---[ Alternative generator ]-------------------------------------------------

GeneratorBase class
end-class Forth-Alt-Generator

: generate-alternative1  ( -- )
   operand1 first-set-test
   save-if
   operand1 generate
   save-else
   operand2 generate
   save-then
;

: generate-alternative2  ( -- )
   operand1 first-set-test save-0=
   operand2 first-set-test save-and
   save-if
   operand2 generate
   save-else
   operand1 generate
   save-then
;

:noname  ( gen-class -- )
   get-section >r nonterm-buf
   drop operand1 compute
   if
      generate-alternative2
   else
      generate-alternative1
   then
   drop
   r> restore-section
; Forth-Alt-Generator defines generate()

Forth-Alt-Generator Alt-Syntax-Expr set-generator

\ ---[ Option generator ]------------------------------------------------------

GeneratorBase class
end-class Forth-Option-Generator

:noname  ( gen-class -- )
   get-section >r nonterm-buf
   drop operand first-set-test
   save-if
   operand generate
   save-then
   r> restore-section
; Forth-Option-Generator defines generate()

Forth-Option-Generator Option-Syntax-Expr set-generator

\ ---[ *Repetition generator ]-------------------------------------------------

GeneratorBase class
end-class Forth-*Repetition-Generator

:noname  ( gen-class -- )
   get-section >r nonterm-buf
   drop save-begin
   operand first-set-test
   save-while
   operand generate
   save-repeat
   r> restore-section
; Forth-*Repetition-Generator defines generate()

Forth-*Repetition-Generator *Repetition-Syntax-Expr set-generator

\ ---[ +Repetition generator ]-------------------------------------------------

GeneratorBase class
end-class Forth-+Repetition-Generator

:noname  ( gen-class -- )
   get-section >r nonterm-buf
   drop
   save-begin
   operand generate
   operand first-set-test
   save-0= save-until
   r> restore-section
; Forth-+Repetition-Generator defines generate()

Forth-+Repetition-Generator +Repetition-Syntax-Expr set-generator

\ ---[ Action generator ]------------------------------------------------------

\ Actions are saved as text as the grammar is loaded. At code generation
\ the actions are copied to the appropriate place in non-terminal
\ definitions

ActionGeneratorBase class
end-class Forth-Action-Generator

\ This class is for Gray format input where actions are {{ ... }}
\ A sequence of actions in Forth cannot contain a word with a name
\ containing }} e.g. foo}}bar as the }} will stop the copying of the action

\ Actions are copied as is because actions must not be split across lines
\ e.g. s" blah blah blah" must be saved on one line

:noname  ( gen-class -- )
   get-section >r nonterm-buf
   drop 0 action                       ( -- f ad )
   begin
      dup c@
   while
      over if save-newline then
      nip -1 swap          
      dup count tuck                   ( -- ad u caddr u )
      save-string                      ( -- ad u )
      1+ chars +                       ( -- ad' )
   repeat
   2drop
   r> restore-section
; Forth-Action-Generator defines generate()

\ In copyActions if the delimiting string (caddr1 u1) is not found within a
\ line then >IN is not set to the end of the line. This does not matter as
\ REFILL is called before another search is undertaken

: copyActions  ( syntax-expr ad syntax-expr gen-class caddr1 u1 -- syntax-expr )
   get-section >r action-buf 2>r    ( -- syntax-expr ad syntax-expr gen-class )
   drop new-context here swap !     ( -- syntax-expr )
   begin
      source >in @ /string          ( -- syntax-expr caddr1 u1 )
      2dup 2r@ search               ( -- syntax-expr caddr1 u1 caddr2 u2 f )
      >r swap source drop - >in !   ( -- syntax-expr caddr1 u1 u2 f )
      r@ if - else drop then        ( -- syntax-expr caddr1 u3 )
      -trailing string>buf r> 0=    ( -- syntax-expr f2 )
   while
      refill 0=
      abort" Actions: unexpected end of file"
   repeat
   0 c,           \ Mark end of action with 0 length
   2r> 2drop r> restore-section
;

:noname  ( syntax-expr ad syntax-expr gen-class -- syntax-expr )
   s" }}" copyActions
; Forth-Action-Generator defines begin-action

\ end-action will be called as begin-action does not consume }}

:noname  ( syntax-expr class -- syntax-expr )
   drop old-context
; Forth-Action-Generator defines end-action

Forth-Action-Generator Action-Syntax-Expr set-generator

\ This class is for conventional BNF input where actions are {...}

Forth-Action-Generator class
end-class Forth-BNF-Action-Generator

\ This action scanner does not consume the terminating } as that has to be
\ recognised by the BNF parser

:noname  ( syntax-expr ad syntax-expr gen-class -- syntax-expr )
   s" }" copyActions
; Forth-BNF-Action-Generator defines begin-action

\ Method end-action should never be called

' internal-error Forth-BNF-Action-Generator defines end-action

Forth-BNF-Action-Generator BNF-Action-Syntax-Expr set-generator

\ ---[ Non-terminal generator ]------------------------------------------------

Pass2GeneratorBase class
end-class Forth-nt-Generator

\ Note that defer is written directly to the output file as we may be in the
\ middle of saving a rule to the non-terminal buffer

: write-defer-name     \ For forward references to non-terminals
   writeNewLine s" defer " writeText nt-name 2@ writeLine
;

:noname  ( gen-class -- )
   get-section >r nonterm-buf
   drop exec @ deferred @ or 0=
   if
      write-defer-name
      true deferred !
   then
   exec @ nt-name 2@ current-nt 2@ d= and
   if save-recurse else nt-name 2@ save-string then
   r> restore-section
; Forth-nt-Generator defines generate()

:noname  ( f class -- )
   [ GeneratorBase :: begin-def ]
   get-section >r nonterm-buf
    here exec !
   r> restore-section
; Forth-nt-Generator defines begin-def


:noname  ( class -- )    \ generates the code for a rule
   get-section >r nonterm-buf
   drop 0 org here exec !
   deferred @
   if
      save-:noname
   else
      save-: nt-name 2@ save-defname
   then
   nt-name 2@ current-nt 2!
   get-body generate
   save-; deferred @
   if
      save-is nt-name 2@ save-string
   then
   save-newline 0 c, exec @ write-buf
   r> restore-section
; Forth-nt-Generator defines code-nt

Forth-nt-Generator nt-Syntax-Expr set-generator

\ -----------------------------------------------------------------------------
\ To read in tokens as used by LexGen

\ Defines a token that returns a unique value

variable tokenval 1 tokenval !

: token ( -- ) ( use: token name ) ( name: -- n )
   tokenval @ constant
   1 tokenval +!
;

\ -----------------------------------------------------------------------------

.( gracegen.fth loaded ) .s
