\ InputSource.fth  provides an ANS interface to a general text file so
\ that ANS words such as REFILL SOURCE PARSE etc can be used on a text file
\ or a string

\ Status: Beta version, SAVE-INPUT and RESTORE-INPUT to be completed

\ Copyright (C) Gerry Jackson 2008

\ This software is free; you can redistribute it and/or modify it in
\ any way provided you acknowledge the original source and copyright

\ This program is distributed in the hope that it will be useful,
\ but WITHOUT ANY WARRANTY; without even the implied warranty of
\ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
\ ------------------------------------------------------------------------------

0 [if]
It circumvents the problem that input sources can only be opened by
words such as INCLUDED and EVALUATE which also invoke the Forth text
interpreter. Files opened by OPEN-FILE cannot become an input source and
therefore cannot be processed using REFILL etc. Similarly strings cannot be
parsed with PARSE etc.

This is done using a Bridge design pattern to execute either the original
ANS Forth definitions of REFILL etc or a set of equivalent words for strings
or files defined using words from the File-Access word set

When a text file or string is to be used as an input source it will not be
interpreted directly by the Forth system at all i.e. user defined words will do
whatever is necessary to the contents of the file e.g. parse and process it. So
don't expect words such as ( or \ to provide a comment, if comments are needed
the user has to do it.

Caveats:
   1. Initially not redefining ' or ['] to use a string or file source therefore
      don't use them in definitions to parse a file source
   2. Don't use any other parsing words in definitions such as CREATE, CONSTANT
      CHAR etc. They can always be added later

Possible enhancements:
   1. Create a NullSource class and object that an unused file source variable
      or value can be set to, prevents crashes in case of errors.
[then]

.( InputSources.fth loading ...) cr 

\ ---[ Base class ]-------------------------------------------------------------

object class
   1 cells var isbufsize   \ Do not change the order of ...
   1 cells var isbuf       \ ...
   1 cells var isnchars    \ ...
   1 cells var is>in       \ ... these fields, see (refill) etc definitions
   method (>in)
   method (refill)
   method (source)
   method (source-id)
   method (parse)
   method (parse-name)
   method (word)
   method (save-input)
   method (restore-input)
end-class InputSourceBase

0 value current-source
0 value ANS-source
: set-source  ( fs -- ) to current-source ;
: set-ANS-source  ( -- )  ANS-source set-source ;

:noname  ( is -- )
   dup current-source =
   if set-ANS-source then           \ Don't leave current-source dangling
   [ object :: delete ]
; InputSourceBase defines delete

\ ---[ ANS Forth versions ]-----------------------------------------------------

InputSourceBase class
end-class ANSInputSource

ANSInputSource new to ANS-source set-ANS-source

:noname  ( is -- )
   drop                 \ Cannot delete this
; ANSInputSource defines delete

\ These methods simply call ANS Forth definitions

:noname drop >in ; ANSInputSource defines (>in)
:noname drop refill ; ANSInputSource defines (refill)
:noname drop source ; ANSInputSource defines (source)
:noname drop source-id ; ANSInputSource defines (source-id)
:noname drop parse ; ANSInputSource defines (parse)
:noname drop parse-name ; ANSInputSource defines (parse-name)
:noname drop word ; ANSInputSource defines (word)
:noname drop save-input ; ANSInputSource defines (save-input)
:noname drop restore-input ; ANSInputSource defines (restore-input)

\ ---[ File-Access word set versions ]------------------------------------------

InputSourceBase class
   1 cells var isid
end-class FileSource

128 constant fbufsize      \ Default buffer size

:noname  ( fileid class -- fs )
   [ InputSourceBase :: new ]             ( -- fileid fs )
   >r fbufsize 2 + chars allocate throw   ( -- fileid ad )
   r@ isbuf ! r@ isid !                   ( -- )
   fbufsize r@ isbufsize !
   0 r@ isnchars ! 0 r@ is>in ! r>        ( -- fs ) 
; FileSource defines new

\ Method delete does not close the file, that is the user's responsibility,
\ delete merely frees the buffer and deletes the file source object

:noname  ( fs -- )
   dup isbuf @ free throw
   [ InputSourceBase :: delete ]
; FileSource defines delete

:noname  ( fs -- ad )  is>in ; FileSource defines (>in)

:noname  ( fs -- flag )
   >r 0 r@ isbufsize 2@ r@ isid @ read-line     ( -- 0 u f ior )
   swap 0= or over r@ isbufsize @ = or          ( -- 0 u f2 )
   if
      r> 2drop                                  ( -- false )
   else
      r> isnchars 2! true                       ( -- true )
   then
; FileSource defines (refill)

:noname  ( fs -- caddr u )  isbuf 2@ swap ; FileSource defines (source)

:noname  ( fs -- fileid )  isid @ ; FileSource defines (source-id)

\ With a suitable choice of xt scan-chars can be used to skip delimiters
\ or scan until a delimiter has been read. 

: scan-chars  ( end start delim xt -- end ad )
   2>r                  ( -- end start )
   begin
      2dup >
   while
      dup c@ 2r@        ( -- end start ch delim xt )
      execute           ( -- end start flag )
   while
      char+             ( -- end ad )
   repeat then
   2r> 2drop
;

: do-parse  ( fs delim flag -- ad u )
   2>r dup >r                    ( -- fs )
   dup (source) chars over       ( -- fs ad u ad )
   + swap                        ( -- fs end ad )
   r> is>in @ + r>               ( -- fs end start flag )
   if                            \ Skip leading delimiters
      r@ dup bl =                ( -- fs end start delim flag2 )
      if ['] <= else ['] = then  ( -- fs end start delim xt )
      scan-chars                 ( -- fs end start' )
   then
   r> over >r dup bl =           ( -- fs end start' delim )
   if ['] > else ['] <> then     ( -- fs end start' delim xt )
   scan-chars                    ( -- fs end ad )     \ Scan for delimiter
   nip swap >r                   ( -- ad )
   dup char+ r@ (source) >r -    ( -- ad n )
   r> min r> is>in !             \ Ensure is>in <= number of characters )
   r> swap over -                ( -- ad u )
;

:noname  ( char fs "ccc<char>" -- caddr u )
   swap false do-parse
; FileSource defines (parse)

:noname  ( fs "spaces<name>" -- caddr u )
   bl true do-parse
; FileSource defines (parse-name)

:noname  ( char fs "<chars>ccc<char>" -- c-addr )
   swap true do-parse dup     ( -- caddr1 u1 u1 )
   [ s" max-char" environment?
     0= [if] 255 [then] ]
   literal >                     \ Is string too long to be counted?
   over 2 chars + unused > or    \ Is there sufficient dataspace for it?
   -18 and throw                 \ "Parsed string overflow"
   dup >r here c!                \ Save string length
   here char+ r@ move            \ Save string
   bl here r> + char+ c!         \ Save blank at end
   here                          ( -- c-addr )
; FileSource defines (word)

\ ---[ String source class ]----------------------------------------------------
\ Inherits from FileSource

FileSource class
end-class StringSource

:noname  ( c-addr u class -- ss )   \ ss = string source object
   [ object :: new ]             ( -- c-addr u ss )
   >r dup r@ isbufsize !
   r@ isnchars !
   r@ isbuf !
   0 r@ is>in !
   -1 r@ isid ! r>               ( -- ss ) 
; StringSource defines new

:noname  ( fs -- )
   [ InputSourceBase :: delete ]
; StringSource defines delete

:noname  ( ss -- false ) drop false ; StringSource defines (refill)

\ ---[ Redefinition of standard words to use class methods ]--------------------

: >in  ( -- ad )  current-source (>in) ;
: refill  ( -- flag )     current-source (refill) ;
: source  ( -- caddr u )  current-source (source) ;
: source-id  ( -- x )     current-source (source-id) ;
: parse  ( char "c*" -- caddr u )  current-source (parse) ;
: parse-name  ( "name" -- caddr u )  current-source (parse-name) ;
: word  ( char "c*" -- caddr )  current-source (word) ;
: save-input  ( -- x1...xn n )  current-source (save-input) ;
: restore-input ( x1...xn n -- )  current-source (restore-input) ; 

\ ---[ FILE-SOURCE, STRING-SOURCE and CLOSE-SOURCE ]----------------------------

: file-source    ( fileid -- fs )   FileSource new ;
: string-source  ( caddr u -- ss )  StringSource new ;
: close-source   ( is -- ) delete ;

\ ------------------------------------------------------------------------------

.( InputSources.fth loaded. ) .s
