\ Lexical Scanner for state transition tables generated by LexGen

\ **** Note that the state transition table must be included before this file

\ -----------------------------------------------------------------------------
\ Copyright (C) Gerry Jackson 2008

\ This software is free; you can redistribute it and/or modify it in
\ any way provided you acknowledge the original source and copyright

\ This program is distributed in the hope that it will be useful,
\ but WITHOUT ANY WARRANTY; without even the implied warranty of
\ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
\ -----------------------------------------------------------------------------

\ This file requires the following library files to have been loaded:
\     xmini_oof.fth and inputsources.fth

.( lexscanner2_0.fth loading ...) cr

\ ---[ Versions ]---------------------------------------------------------------

\ 2.0 - Rewritten assuming that the library file InputSources.fth is in use.
\       This allows a file to be read and parsed etc using standard words such
\       as REFILL PARSE etc. In this case the lexical scanner can obtain the
\       line to be scanned via SOURCE with >IN giving the cursor position
\     - LexScanner class introduced to make the lexical scanner reentrant so
\       that multiple scanners may be used in a single application
\ 1.1 - Now uses library file Files.fth to open and close the file to be lexed
\     - Words getRestOfLine amd getCursorChar added
\     - getLine inserts a 0 byte just beyond the end of the line
\     - cursor and endOfLine initialised to an empty line

\ ---[ User interface ]---------------------------------------------------------
\
\ There are four words in the user interface:
\
\     init-scanner  ( ad1 n1 ad2 n2 -- scanner )
\               where ad1 = address of Base/Default array
\                     n1  = number of states
\                     ad2 = address of Check/Next array
\                     n2  = number of Check/Next entries
\     next-token    ( scanner -- c-addr u token )
\     close-scanner ( scanner -- )
\     token         ( "spaces<name>" -- )  <name> execution ( -- n )
\
\ In addition the user must have opened the file to be scanned and made it a
\ file source by making use of library file inputsources.fth and loaded the
\ first line using REFILL. Tokens can then be recognised (or not) by the
\ scanner by calling next-token until the end of the file.
\ close-scanner may be called when scanning is complete.
\
\ The lexical scanner is fully reentrant so that an application can use
\ multiple scanners
\
\ The lexical scanner relies on the fact that the state transition tables (STT)
\ generated by LexGen will have loaded four VALUEs:
\     #states, BaseDefaultData, maxCheck and CheckNextData
\ These are saved in a LexScanner object when init-scanner is called.
\ Hence when multiple scanners are needed, init-scanner must be called after
\ a STT is loaded and before the next STT is loaded
\
\ The word token is provided for convenience to define token constants and, by
\ using the same file as LexGen, keeping them in synchronism with those embedded
\ in the STT. 

\ ---[ LexScanner class ]-------------------------------------------------------

object class
   \ Keep the next two vars in the same order (see 2@ in >addresses)
   1 cells var sc_#states     \ Number of states in the transition table
   1 cells var sc_basedef     \ Base address of the Base/Default array
   \ Keep the next two vars in the same order (see 2@ in >addresses & nextState)
   1 cells var sc_maxcheck    \ Upper address of the Check/Next array
   1 cells var sc_checknext   \ Base address of the Check/Next array
end-class LexScanner

:noname  ( ad1 n1 ad2 n2 class -- scan )
   [ object :: new ]       ( -- ad1 n1 ad2 n2 scan )
   dup >r sc_maxcheck 2!   \ Note >addresses converts sc_maxcheck to an address
   r@ sc_#states  2! r>
; LexScanner defines new

\ ---[ To access lex arrays ]---------------------------------------------------

1 cells constant 1cell
2 cells constant 2cells
3 cells constant 3cells

: BaseDefault  ( index scan -- ad )
   sc_basedef @ swap 3cells * +
;

: CheckNext    ( index scan -- ad )
   sc_checknext @ swap 2cells * +
;

\ These four definitions are for readability in the scanner

: lexBase ; immediate         \ Compiles nothing
: lexDefault postpone cell+ ; immediate
: lexToken postpone cell+ postpone cell+ ; immediate 
: lexCheck ; immediate
: lexNext postpone cell+ ; immediate

\ ---[ Conversion of lex array data to absolute addresses ]---------------------
\ Using absolute addresses is faster than using the array data to index the
\ arrays. An alternative is to replace the , in the LexTables data with a word
\ that does the conversion as the file is read

: ?invalid  ( ad1 scan -- ad2 | 0 )   \ 0 if contents of ad1 negative
   swap @ dup 0<
   if
      2drop 0
   else
      swap BaseDefault
   then
;

: >addresses  ( scan -- )
   dup sc_#states 2@ 3cells * over + swap
   ?do
      i lexBase @ over CheckNext i lexBase !
      i lexDefault over ?invalid i lexDefault !
      3cells
   +loop
   dup sc_maxCheck 2@ 2cells * over + swap
   ?do
      i lexCheck over ?invalid i lexCheck !
      i lexNext  over ?invalid i lexNext !
      2cells
   +loop
   dup sc_maxcheck @ over CheckNext swap sc_maxcheck !
;

: init-scanner  ( ad1 n1 ad2 n2 -- scan )
   LexScanner new dup >addresses
;

: close-scanner  ( scan -- )  delete ;

\ ---[ Lexical scanner ]--------------------------------------------------------
\  The user should call nextToken

: getChar  ( -- ch )    \ ch = -1 is end of line
   source >in @ 2dup <=             ( -- c-addr1 u1 u2 f )
   if drop 2drop -1 exit then
   nip chars + c@                   ( -- ch )
   1 >in +!
;

: getCursor  ( -- caddr )
   source drop >in @ chars +
;

\ nextState returns 0 if there is no valid next state
\ state and state2 are addesses, ch' is a character converted to cell pairs
\ by the caller

: nextState    ( state ch' scan -- state2 )
   2>r                        ( -- state ) ( R: -- ch' scan )
   dup lexBase @ 2r@ drop +   ( -- state ad1 )
   dup r@ sc_maxcheck 2@
   within    ( -- state ad1 f ) \ f = 0 is out of range
   if
      2dup lexCheck @ =       ( -- state ad1 f )
      if
         nip lexNext @        ( -- state2 )
         2r> 2drop exit
      then
   then
   drop lexDefault @ dup      ( -- state3 state3 )
   if 2r> recurse exit then   ( -- state4 )
   2r> 2drop                  ( -- state3 )  \ where state3 = 0
;

\ nextToken returns 0 for invalid token, -1 for end of line, else valid token
\ Note when a state has a valid token, this is remembered and we carry on
\ until we can get no further. This allows a longer lexeme to be recognised
\ and backtracking (within the line) if there is none such. 

: next-token    ( scan -- caddr u token )
   >r getCursor dup char+           ( -- caddr1 caddr2 ) ( R: -- scan )
   getChar dup 0<                   ( -- caddr1 caddr2 ch f )
   if r> 2drop -1 exit then         ( -- x x -1 )  \ -1 for end of line
   2cells *                         ( -- caddr1 caddr2 ch' )
   0 r@ sc_basedef @ rot            ( -- caddr1 caddr2 tok state ch' )
   begin
      r@ nextState dup              ( -- caddr1 caddr2 tok state2 state2)
   while
      dup lexToken @ 0>             ( -- caddr1 caddr2 tok state2 f )
      if
         >r 2drop getCursor         ( -- caddr1 caddr3 )
         r@ lexToken @ r>           ( -- caddr1 caddr3 tok2 state2 )
      then
      getChar 2cells *
   repeat
   r> 2drop >r                      ( -- caddr1 caddr3 ) ( R: -- tok )
   2dup source drop - >in !         ( -- caddr1 caddr3 caddr1 )
   - r>                             ( -- caddr1 u token )
;

\ ---[ To read in tokens as used by LexGen ]------------------------------------

\ Defines a token that returns a unique value

variable tokenval 1 tokenval !

: token ( -- ) ( use: token name ) ( name: -- n )
   tokenval @ constant
   1 tokenval +!
;

\ -----------------------------------------------------------------------------

.( lexscanner2_0.fth loaded. ) .s

